/*-
 * %%----------------------------------------------------------------------------------------------
 * Solidify Framework - Solidify Index - ElasticsearchService.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.unige.solidify.index.indexing.elasticsearch;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.stream.Collectors;

import org.elasticsearch.ElasticsearchStatusException;
import org.elasticsearch.action.bulk.BulkRequest;
import org.elasticsearch.action.bulk.BulkResponse;
import org.elasticsearch.action.delete.DeleteRequest;
import org.elasticsearch.action.delete.DeleteResponse;
import org.elasticsearch.action.index.IndexRequest;
import org.elasticsearch.action.index.IndexResponse;
import org.elasticsearch.action.search.SearchRequest;
import org.elasticsearch.action.search.SearchResponse;
import org.elasticsearch.action.update.UpdateRequest;
import org.elasticsearch.action.update.UpdateResponse;
import org.elasticsearch.client.RequestOptions;
import org.elasticsearch.client.RestHighLevelClient;
import org.elasticsearch.client.indices.GetIndexRequest;
import org.elasticsearch.index.query.AbstractQueryBuilder;
import org.elasticsearch.index.query.BoolQueryBuilder;
import org.elasticsearch.index.query.QueryBuilder;
import org.elasticsearch.index.query.QueryBuilders;
import org.elasticsearch.index.query.QueryStringQueryBuilder;
import org.elasticsearch.index.query.RangeQueryBuilder;
import org.elasticsearch.rest.RestStatus;
import org.elasticsearch.search.SearchHit;
import org.elasticsearch.search.SearchHits;
import org.elasticsearch.search.aggregations.Aggregation;
import org.elasticsearch.search.aggregations.AggregationBuilders;
import org.elasticsearch.search.aggregations.Aggregations;
import org.elasticsearch.search.aggregations.bucket.terms.Terms;
import org.elasticsearch.search.aggregations.bucket.terms.TermsAggregationBuilder;
import org.elasticsearch.search.builder.SearchSourceBuilder;
import org.elasticsearch.search.sort.SortOrder;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.Pageable;

import ch.unige.solidify.IndexConstants;
import ch.unige.solidify.IndexProperties;
import ch.unige.solidify.SolidifyConstants;
import ch.unige.solidify.exception.SolidifyRuntimeException;
import ch.unige.solidify.exception.SolidifyValidationException;
import ch.unige.solidify.index.ElasticsearchClientProvider;
import ch.unige.solidify.index.indexing.IndexingService;
import ch.unige.solidify.model.index.IndexMetadata;
import ch.unige.solidify.rest.BooleanClauseType;
import ch.unige.solidify.rest.FacetPage;
import ch.unige.solidify.rest.FacetRequest;
import ch.unige.solidify.rest.FacetResult;
import ch.unige.solidify.rest.FacetResultValue;
import ch.unige.solidify.rest.RestCollectionPage;
import ch.unige.solidify.rest.SearchCondition;
import ch.unige.solidify.rest.SearchConditionType;
import ch.unige.solidify.util.JSONTool;
import ch.unige.solidify.util.StringTool;
import ch.unige.solidify.validation.ValidationError;

public abstract class ElasticsearchService<T extends IndexMetadata> extends IndexingService<T> {
  private static final Logger log = LoggerFactory.getLogger(ElasticsearchService.class);

  private static final int NB_RETRY_ON_CONFLICT = 5;
  private final RestHighLevelClient esClt;

  protected ElasticsearchService(IndexProperties indexProperties, ElasticsearchClientProvider elasticsearchClientProvider) {
    super(indexProperties);
    this.esClt = elasticsearchClientProvider.getElasticsearchClient();
  }

  @Override
  public boolean delete(String indexName, T t) {
    final DeleteResponse res;
    try {
      res = this.esClt.delete(new DeleteRequest(t.getIndex(), t.getResId()), RequestOptions.DEFAULT);
    } catch (IOException e) {
      throw new SolidifyRuntimeException(
              "Error when deleting archive " + t.getResId() + " from index " + t.getIndex(), e);
    }
    return (res.status() == RestStatus.OK);
  }

  @Override
  public void deleteList(String indexName, List<String> ids) {
    final RequestOptions refreshOption = RequestOptions.DEFAULT.toBuilder().addParameter("refresh", "true").build();
    final BulkRequest bulkRequest = new BulkRequest();
    for (String id : ids) {
      bulkRequest.add(new DeleteRequest(indexName, id));
    }
    try {
      BulkResponse bulkResponse = this.esClt.bulk(bulkRequest, refreshOption);
      RestStatus restStatus = bulkResponse.status();
      if (!restStatus.equals(RestStatus.OK)) {
        throw new SolidifyRuntimeException("Failure when deleting list of archive, status is " + restStatus);
      }
    } catch (IOException e) {
      throw new SolidifyRuntimeException("IOException when deleting list of archives", e);
    }
  }

  @Override
  public List<T> findAll(String indexName, T search) {
    final List<T> list = new ArrayList<>();
    final SearchSourceBuilder searchSourceBuilder = new SearchSourceBuilder()
            .query(QueryBuilders.matchAllQuery())
            .size(RestCollectionPage.DEFAULT_SIZE_PAGE);
    final SearchRequest searchRequest = new SearchRequest(indexName.split(SolidifyConstants.FIELD_SEP))
            .source(searchSourceBuilder);
    try {
      final SearchResponse res = this.esClt.search(searchRequest, RequestOptions.DEFAULT);
      for (final SearchHit h : res.getHits().getHits()) {
        list.add(this.convertHit(h));
      }
      return list;
    } catch (IOException e) {
      throw new SolidifyRuntimeException("Error when finding all from index " + indexName, e);
    }
  }

  @Override
  public FacetPage<T> findAll(String indexName, T search, Pageable pageable) {
    final SearchSourceBuilder searchSourceBuilder = new SearchSourceBuilder()
            .query(QueryBuilders.matchAllQuery());
    this.setPageAndSort(searchSourceBuilder, pageable);
    final SearchRequest searchRequest = new SearchRequest(indexName.split(SolidifyConstants.FIELD_SEP))
            .source(searchSourceBuilder);
    try {
      final SearchResponse response = this.esClt.search(searchRequest, RequestOptions.DEFAULT);
      return this.getPage(response, pageable);
    } catch (IOException e) {
      throw new SolidifyRuntimeException("Error when finding all from index " + indexName, e);
    }
  }

  @Override
  public T findOne(String indexName, String id) {
    final SearchSourceBuilder searchSourceBuilder = new SearchSourceBuilder()
            .query(QueryBuilders.termQuery("_id", id));
    final SearchRequest searchRequest = new SearchRequest(indexName.split(SolidifyConstants.FIELD_SEP))
            .source(searchSourceBuilder);
    try {
      final SearchResponse response = this.esClt.search(searchRequest, RequestOptions.DEFAULT);
      if (response.getHits().getHits().length == 0) {
        return null;
      } else if (response.getHits().getHits().length == 1) {
        return this.convertHit(response.getHits().getHits()[0]);
      } else {
        return this.mergeHits(response.getHits().getHits());
      }
    } catch (IOException e) {
      throw new SolidifyRuntimeException("Error when searching " + id + " from index " + indexName, e);
    }
  }

  @Override
  public T save(String indexName, T metadata) {
    this.checkArchive(metadata);
    this.checkIndex(metadata.getIndex());
    final IndexRequest request = new IndexRequest(metadata.getIndex())
            .id(metadata.getResId())
            .source(metadata.getMetadata());

    try {
      final IndexResponse res = this.esClt.index(request, RequestOptions.DEFAULT);
      if (res.status() == RestStatus.CREATED || res.status() == RestStatus.OK) {
        return metadata;
      } else {
        throw new SolidifyRuntimeException(
                "Problem in creating index (" + metadata.getResId() + "," + metadata.getMetadata() + "," + res.status() + ")");
      }
    } catch (IOException e) {
      throw new SolidifyRuntimeException("Error when saving " + metadata.getResId() + " in index " + metadata.getIndex(), e);
    }
  }

  @Override
  public Page<T> search(String indexName, String query, Pageable pageable) {
    final QueryStringQueryBuilder q = QueryBuilders.queryStringQuery(query);
    return this.search(indexName, q, Collections.emptyList(), pageable);
  }

  @Override
  public Page<T> search(String indexName, String query, String setQuery, String field, String from, String to, Pageable pageable) {
    return this.search(indexName, query, Collections.emptyList(), setQuery, field, from, to, null, pageable);
  }

  @Override
  public FacetPage<T> search(String indexName, String query, List<FacetRequest> facetRequests, String setQuery, String field,
          String from, String to, String filter, Pageable pageable) {

    /*
     * Filtering can combine 3 logical parts:
     *
     * - a given query string - a set - a range on a field
     *
     * They are all combined by using a BoolQuery
     */
    final BoolQueryBuilder boolQuery = QueryBuilders.boolQuery();
    final List<QueryBuilder> mustQueries = boolQuery.must();

    /*
     * given query criteria
     */
    if (!StringTool.isNullOrEmpty(query)) {
      final QueryStringQueryBuilder q = QueryBuilders.queryStringQuery(query);
      mustQueries.add(q);
    }

    /*
     * set criteria
     */
    if (!StringTool.isNullOrEmpty(setQuery)) {
      final QueryStringQueryBuilder q = QueryBuilders.queryStringQuery(setQuery);
      mustQueries.add(q);
    }

    /*
     * range criteria
     */
    if (!StringTool.isNullOrEmpty(field) && (!StringTool.isNullOrEmpty(from) || !StringTool.isNullOrEmpty(to))) {
      RangeQueryBuilder q = QueryBuilders.rangeQuery(field);

      if (!StringTool.isNullOrEmpty(from)) {
        q = q.from(from).includeLower(true);
      }
      if (!StringTool.isNullOrEmpty(to)) {
        q = q.to(to).includeUpper(true);
      }
      mustQueries.add(q);
    }

    if (!StringTool.isNullOrEmpty(filter)) {
      final QueryStringQueryBuilder q = QueryBuilders.queryStringQuery(filter);
      mustQueries.add(q);
    }

    return this.search(indexName, boolQuery, facetRequests, pageable);
  }

  @Override
  public Page<T> searchRange(String indexName, String field, String from, String to, Pageable pageable) {
    RangeQueryBuilder q = QueryBuilders.rangeQuery(field);
    if (!StringTool.isNullOrEmpty(from)) {
      q = q.from(from).includeLower(true);
    }
    if (!StringTool.isNullOrEmpty(to)) {
      q = q.to(to).includeUpper(true);
    }
    return this.search(indexName, q, Collections.emptyList(), pageable);
  }

  @Override
  public FacetPage<T> search(String indexName, List<SearchCondition> conditions, List<FacetRequest> facetRequests,
          Pageable pageable) {

    final BoolQueryBuilder boolQuery = this.getBooleanQueryBuilder(conditions);

    return this.search(indexName, boolQuery, facetRequests, pageable);
  }

  private BoolQueryBuilder getBooleanQueryBuilder(List<SearchCondition> conditions) {
    final BoolQueryBuilder boolQuery = QueryBuilders.boolQuery();
    this.fillBuilders(boolQuery, conditions, BooleanClauseType.MUST);
    this.fillBuilders(boolQuery, conditions, BooleanClauseType.FILTER);
    this.fillBuilders(boolQuery, conditions, BooleanClauseType.MUST_NOT);
    this.fillBuilders(boolQuery, conditions, BooleanClauseType.SHOULD);
    return boolQuery;
  }

  private void fillBuilders(BoolQueryBuilder boolQuery, List<SearchCondition> conditions, BooleanClauseType booleanClauseType) {
    List<QueryBuilder> queryBuilders;
    switch (booleanClauseType) {
      case MUST:
        queryBuilders = boolQuery.must();
        break;
      case FILTER:
        queryBuilders = boolQuery.filter();
        break;
      case MUST_NOT:
        queryBuilders = boolQuery.mustNot();
        break;
      case SHOULD:
        queryBuilders = boolQuery.should();
        break;
      default:
        throw new SolidifyRuntimeException("Unknown BooleanClauseType value");
    }

    this.fillBuilders(queryBuilders,
            conditions.stream().filter(condition -> condition.getBooleanClauseType() == booleanClauseType).collect(Collectors.toList()));
  }

  private void fillBuilders(List<QueryBuilder> builders, List<SearchCondition> conditions) {
    for (SearchCondition condition : conditions) {
      QueryBuilder builder;
      if (condition.getType() == SearchConditionType.MATCH) {
        builder = this.getMatchQueryBuilder(condition);
      } else if (condition.getType() == SearchConditionType.TERM) {
        builder = this.getTermQueryBuilder(condition);
      } else if (condition.getType() == SearchConditionType.RANGE) {
        builder = this.getRangeQueryBuilder(condition);
      } else if (condition.getType() == SearchConditionType.QUERY) {
        builder = this.getQueryQueryBuilder(condition);
      } else if (condition.getType() == SearchConditionType.SIMPLE_QUERY) {
        builder = this.getSimpleQueryQueryBuilder(condition);
      } else if (condition.getType() == SearchConditionType.NESTED_BOOLEAN) {
        builder = this.getNestedBooleanQueryBuilder(condition);
      } else {
        throw new SolidifyRuntimeException("SearchConditionType '" + condition.getType() + "' not implemented");
      }
      builders.add(builder);
    }
  }

  private QueryBuilder getMatchQueryBuilder(SearchCondition condition) {
    if (!StringTool.isNullOrEmpty(condition.getField()) && !StringTool.isNullOrEmpty(condition.getValue())) {
      // 'match' query
      return QueryBuilders.matchQuery(condition.getField(), condition.getValue());
    } else if (!condition.getMultiMatchFields().isEmpty() && !StringTool.isNullOrEmpty(condition.getValue())) {
      // 'multi_match' query
      return QueryBuilders.multiMatchQuery(condition.getValue(),
              condition.getMultiMatchFields().toArray(new String[condition.getMultiMatchFields().size()]));
    } else {
      throw new SolidifyRuntimeException("Invalid 'match' search condition");
    }
  }

  private QueryBuilder getTermQueryBuilder(SearchCondition condition) {
    if (!StringTool.isNullOrEmpty(condition.getField()) && !StringTool.isNullOrEmpty(condition.getValue())) {
      // 'term' query
      return QueryBuilders.termQuery(condition.getField(), condition.getValue());
    } else if (!StringTool.isNullOrEmpty(condition.getField()) && !condition.getTerms().isEmpty()) {
      // 'terms' query
      return QueryBuilders.termsQuery(condition.getField(), condition.getTerms());
    } else {
      throw new SolidifyRuntimeException("Invalid 'term' search condition");
    }
  }

  private QueryBuilder getRangeQueryBuilder(SearchCondition condition) {
    if (StringTool.isNullOrEmpty(condition.getLowerValue()) && StringTool.isNullOrEmpty(condition.getUpperValue())) {
      throw new SolidifyRuntimeException("Invalid 'range' search condition");
    }

    RangeQueryBuilder rangeQueryBuilder = QueryBuilders.rangeQuery(condition.getField());
    if (!StringTool.isNullOrEmpty(condition.getLowerValue())) {
      rangeQueryBuilder = rangeQueryBuilder.from(condition.getLowerValue()).includeLower(true);
    }
    if (!StringTool.isNullOrEmpty(condition.getUpperValue())) {
      rangeQueryBuilder = rangeQueryBuilder.to(condition.getUpperValue()).includeUpper(true);
    }
    return rangeQueryBuilder;
  }

  private QueryBuilder getQueryQueryBuilder(SearchCondition condition) {
    if (StringTool.isNullOrEmpty(condition.getValue())) {
      throw new SolidifyRuntimeException("Invalid 'query' search condition");
    }
    return QueryBuilders.queryStringQuery(condition.getValue());
  }

  private QueryBuilder getSimpleQueryQueryBuilder(SearchCondition condition) {
    if (StringTool.isNullOrEmpty(condition.getValue())) {
      throw new SolidifyRuntimeException("Invalid 'simple_query' search condition");
    }
    return QueryBuilders.simpleQueryStringQuery(condition.getValue());
  }

  private QueryBuilder getNestedBooleanQueryBuilder(SearchCondition condition) {
    if (condition.getNestedConditions().isEmpty()) {
      throw new SolidifyRuntimeException("Invalid nested boolean condition");
    }

    return this.getBooleanQueryBuilder(condition.getNestedConditions());
  }

  @Override
  public T update(String indexName, T metadata) {
    this.checkArchive(metadata);
    final UpdateRequest request = new UpdateRequest(metadata.getIndex(), metadata.getResId())
            .retryOnConflict(NB_RETRY_ON_CONFLICT)
            .doc(metadata.getMetadata());
    try {
      final UpdateResponse res = this.esClt.update(request, RequestOptions.DEFAULT);
      if (res.status() == RestStatus.OK) {
        return metadata;
      }
      return null;
    } catch (IOException e) {
      throw new SolidifyRuntimeException("Error when updating " + metadata.getResId() + " from index " + metadata.getIndex(), e);
    }
  }

  private void checkArchive(T t) {
    if (!t.check()) {
      final String msg = "Missing fields in " + t;
      log.error(msg);
      throw new SolidifyRuntimeException(msg);
    }
  }

  @Override
  protected void checkIndex(String indexName) {
    super.checkIndex(indexName);
    GetIndexRequest request = new GetIndexRequest(indexName.split(SolidifyConstants.FIELD_SEP));
    try {
      this.esClt.indices().exists(request, RequestOptions.DEFAULT);
    } catch (IOException e) {
      throw new SolidifyRuntimeException("Error when checking index " + indexName, e);
    }
  }

  private T convertHit(SearchHit hit) {
    final T metadata = this.getMetadataObject();
    metadata.setResId(hit.getId());
    metadata.setIndex(hit.getIndex());
    metadata.setMetadata(hit.getSourceAsString());
    return metadata;
  }

  private FacetPage<T> getPage(SearchResponse searchResponse, Pageable pageable) {
    SearchHits searchHits = searchResponse.getHits();
    final List<T> list = new ArrayList<>();
    for (final SearchHit h : searchHits.getHits()) {
      list.add(this.convertHit(h));
    }

    List<FacetResult> facetResults = new ArrayList<>();
    Aggregations aggregations = searchResponse.getAggregations();
    if (aggregations != null) {
      for (Aggregation aggregation : aggregations.asList()) {
        FacetResult facetResult = new FacetResult(aggregation.getName());
        Terms terms = aggregations.get(aggregation.getName());
        List<Terms.Bucket> buckets = (List<Terms.Bucket>) terms.getBuckets();
        for (Terms.Bucket bucket : buckets) {
          facetResult.getValues().add(new FacetResultValue(bucket.getKeyAsString(), bucket.getDocCount()));
          log.trace("Facet '{}' : {} -> {}", aggregation.getName(), bucket.getKeyAsString(), bucket.getDocCount());
        }
        facetResults.add(facetResult);
      }
    }

    return new FacetPage<>(list, facetResults, pageable, searchHits.getTotalHits().value);
  }

  private T mergeHits(SearchHit[] hits) {
    final T metadata = this.getMetadataObject();
    metadata.setResId(hits[0].getId());
    for (final SearchHit hit : hits) {
      if (metadata.getIndex() == null) {
        metadata.setIndex(hit.getIndex());
      } else {
        metadata.setIndex(metadata.getIndex() + SolidifyConstants.FIELD_SEP + hit.getIndex());
      }
      metadata.getMetadata().putAll(JSONTool.convertJson2Map(hit.getSourceAsString()));
    }
    return metadata;
  }

  private FacetPage<T> search(String indexName, AbstractQueryBuilder<?> q, List<FacetRequest> facetRequests, Pageable pageable) {
    final SearchSourceBuilder searchSourceBuilder = new SearchSourceBuilder().query(q);
    final SearchRequest searchRequest = new SearchRequest(indexName.split(SolidifyConstants.FIELD_SEP)).source(searchSourceBuilder);

    // Aggregations
    if (facetRequests != null && !facetRequests.isEmpty()) {
      for (FacetRequest facetRequest : facetRequests) {
        TermsAggregationBuilder aggregation = AggregationBuilders.terms(facetRequest.getName());
        aggregation.field(facetRequest.getField());
        if (facetRequest.getMinCount() != null) {
          aggregation.minDocCount(facetRequest.getMinCount());
        }
        if (facetRequest.getLimit() != null) {
          aggregation.size(facetRequest.getLimit());
        }
        searchSourceBuilder.aggregation(aggregation);
      }
    }

    // Add page & sort parameters
    this.setPageAndSort(searchSourceBuilder, pageable);
    // Execute query
    try {
      final SearchResponse response = this.esClt.search(searchRequest, RequestOptions.DEFAULT);
      FacetPage<T> page = this.getPage(response, pageable);

      // set facets fields names as they are not included in ElasticSearch response
      if (!page.getFacetResults().isEmpty() && facetRequests != null) {
        for (FacetResult result : page.getFacetResults()) {
          for (FacetRequest facetRequest : facetRequests) {
            if (!StringTool.isNullOrEmpty(facetRequest.getName()) && facetRequest.getName().equals(result.getName())) {
              result.setField(facetRequest.getField());
              break;
            }
          }
        }
        // sort facets results as they were requested (Elasticsearch does not necessarily return facets in the same order as requested)
        this.sortFacetResults(page.getFacetResults(), facetRequests);
      }

      return page;
    } catch (IOException e) {
      throw new SolidifyRuntimeException("Error when searching index " + indexName + " with request " + q, e);
    } catch (ElasticsearchStatusException e) {
      if (RestStatus.BAD_REQUEST.equals(e.status())) {
        throw new SolidifyValidationException(new ValidationError(IndexConstants.BAD_SEARCH_REQUEST));
      } else {
        throw e;
      }
    }
  }

  private SearchSourceBuilder setPageAndSort(SearchSourceBuilder searchSourceBuilder, Pageable pageable) {
    // Page Info
    searchSourceBuilder.size(pageable.getPageSize())
            .from(pageable.getPageSize() * pageable.getPageNumber());
    // Check if sort criteria
    if (pageable.getSort().isSorted()) {
      pageable.getSort().forEach(o -> searchSourceBuilder.sort(o.getProperty(), SortOrder.fromString(o.getDirection().toString())));
    }
    return searchSourceBuilder;
  }

  private void sortFacetResults(List<FacetResult> facetResults, List<FacetRequest> sortedFacetRequests) {

    List<String> sortedFacetNames = sortedFacetRequests.stream().map(FacetRequest::getName).collect(Collectors.toList());

    facetResults.sort((left, right) -> {
      int pos1 = sortedFacetNames.indexOf(left.getName());
      int pos2 = sortedFacetNames.indexOf(right.getName());
      return Integer.compare(pos1, pos2);
    });
  }
}
